"""Declares :class:`AssertedSubject`."""
# pylint: disable=unused-argument,no-self-use
from django.db import models
from django.utils.translation import gettext_lazy as _


class AssertedSubject(models.Model):
    """Represents a :term:`Subject` that was asserted by an
    :term:`Identity Provider` or :term:`Asserting Party`.
    """
    USERNAME_FIELD = 'asid'
    REQUIRED_FIELDS = []

    #: Is always ``False``.
    is_anonymous = False

    #: Is always ``True``
    is_authenticated = True

    #: Is always ``True``
    is_active = True

    #: The local :term:`Subject Identifier`. This identifier must only be used
    #: within the scope of the consuming application.
    lsid = models.BigAutoField(
        verbose_name=_("Local Subject ID"),
        primary_key=True,
        blank=False,
        null=False,
        editable=False,
        db_column='lsid'
    )

    #: The identity that was asserted by an :term:`AP`/:term:`IdP` when the
    #: :class:`AssertedSubject` first authenticated with the consuming
    #: application.
    asid = models.CharField(
        verbose_name=_("Asserted Identity"),
        max_length=254,
        blank=False,
        null=False,
        unique=True,
        db_column='asid'
    )

    #: Indicates if the :term:`Asserted Subject` is considered external.
    #: Implementations should ensure that external :term:`Subjects` have
    #: as least privileges as possible, and all privileges that are granted
    #: by default to internal subjects are to be explicitely granted. Defaults
    #: to ``True``.
    external = models.BooleanField(
        verbose_name=_("External"),
        blank=False,
        null=False,
        default=True,
        db_column='is_external'
    )

    #: Indicates if the :term:`Subject` is a :term:`Super User`. Can not be
    #: ``True`` if :attr:`AssertedSubject.external` is also ``True``.
    is_superuser = models.BooleanField(
        verbose_name=_("Superuser"),
        blank=False,
        null=False,
        default=False,
        db_column='is_superuser'
    )

    @property
    def is_staff(self):
        """Return a boolean indicating if the user can access the admin
        pages.
        """
        return self.is_superuser

    @classmethod
    def new(cls):
        """Return a new :class:`AssertedSubject` instance."""
        return cls()

    def get_username(self):
        """Return :attr:`asid`."""
        return self.asid

    def set_password(self, *args, **kwargs):
        """Does nothing."""
        return None

    def check_password(self, *args, **kwargs):
        """Always returns ``False``, since an :term:`Asserted Subject`
        authenticates itself at the :term:`AP`/:term:`IdP`.
        """
        return False

    def set_unusable_password(self, *args, **kwargs):
        """Does nothing, exists for :mod:`django.contrib.auth` compatibility."""
        return None

    def has_perm(self, perm, obj=None):
        """Returns ``True`` if the user has the specified permission, where `perm`
        is in the format ``"<app label>.<permission codename>"``. (see documentation
        on permissions). If the user is inactive, this method will always return
        ``False``. For an active superuser, this method will always return ``True``.

        If `obj` is passed in, this method won’t check for a permission for the model,
        but for this specific object.
        """
        return self.is_superuser

    def has_module_perms(self, package_name):
        """Returns ``True`` if the user has any permissions in the given package
        (the Django app label). If the user is inactive, this method will always
        return ``False``. For an active superuser, this method will always return
        ``True``.
        """
        return self.is_staff or self.is_superuser

    def has_unusable_password(self, *args, **kwargs):
        """Always returns ``True``, since an :term:`Asserted Subject`
        authenticates itself at the :term:`AP`/:term:`IdP`.
        """
        return False

    class Meta: # pylint: disable=missing-class-docstring,too-few-public-methods
        app_label = 'idc'
        constraints = [
            models.CheckConstraint(
                name="assertedsubjects_not_superuser_and_external",
                check=~(models.Q(external=True) & models.Q(is_superuser=True))
            )
        ]
        db_table = 'assertedsubjects'
        default_permissions = []
        verbose_name = _("Asserted Subject")
        verbose_name_plural = _("Asserted Subjects")



"""Declares :class:`ObjectMetaSchema`."""
import marshmallow
import marshmallow.fields


class ObjectMetaSchema(marshmallow.Schema):
    """Declares a schema for `ObjectMetadata` resources."""
    namespace = marshmallow.fields.String(
        required=False,
        default="",
        missing=""
    )

    name = marshmallow.fields.String(
        required=True
    )

    annotations = marshmallow.fields.Dict(
        required=False,
        missing=dict
    )



# pylint: disable=invalid-name,no-self-use
"""Declares :class:`SubjectRepository`."""
from django.apps import apps


class SubjectRepository:
    """Provides an interface to lookup :class:`~iam.ext.idc.models.AssertedSubject`
    objects.
    """
    model_class = 'idc.AssertedSubject'

    @property
    def DoesNotExist(self):
        """Proxy to Django :exc:`DoesNotExist`."""
        return self.model.DoesNotExist

    @property
    def model(self):
        """Return the Django model class used by this repository."""
        return apps.get_model(self.model_class)

    def exists(self, pk):
        """Return a boolean indicating if the entity identified by
        primary key `pk` exists.
        """
        return self.model.objects.filter(pk=pk).exists()

    def get(self, *args, **kwargs):
        """Return a single entity using the provided lookup parameters."""
        return self.model.objects.get(*args, **kwargs)

    def persist(self, dao):
        """Persists a Data Access Object (DAO) to the storage backend."""
        dao.save()



# pylint: skip-file
from .loaders import EnvironmentLoader
from .loaders import SecretsDirectoryLoader
from .registry import BaseIdentityProviverRegistry


class IdentityProvidersRegistry(BaseIdentityProviverRegistry):
    """Maintains a registry of all defined identity providers."""
    pass



# pylint: disable=R0903
"""Declares :class:`EnvironmentLoader`."""
import os

from .base import BaseLoader

from ..oauth2.providers import GoogleOIDCProvider


class EnvironmentLoader(BaseLoader):
    """Loads providers from environment variables."""
    provider_classes = [
        GoogleOIDCProvider
    ]

    def load(self, env=None): # pylint: disable=W0221
        """Invoke the ``fromenv()`` method on :attr:`provider_classes`
        to construct a list of providers.
        """
        providers = []
        for cls in self.provider_classes:
            try:
                providers.append(cls.fromenv(env or os.environ))
            except KeyError:
                continue
        return providers



# pylint: skip-file
import os
import tempfile
import unittest

import ioc
import unimatrix.lib.test
import yaml
from authlib.integrations.django_client import OAuth

from iam.ext.idc.lib import SecretsDirectoryLoader
from iam.ext.idc.lib import IdentityProvidersRegistry


@unimatrix.lib.test.unit
class BaseProviderTestcase(unittest.TestCase):
    __test__ = False

    @classmethod
    def setUpClass(cls):
        cls.secdir = tempfile.mkdtemp()
        os.makedirs(os.path.join(cls.secdir, 'iam.unimatrixone.io/providers'))
        cls.fp = os.path.join(cls.secdir,
            "iam.unimatrixone.io/providers", 'providers.yaml')
        with open(cls.fp, 'w') as f:
            f.write(cls.provider_resource)
        ioc.provide('OAuth2Registry', OAuth(), force=True)

    def setUp(self):
        self.registry = IdentityProvidersRegistry([
            SecretsDirectoryLoader(secdir=self.secdir)
        ])
        self.resource = yaml.safe_load(self.provider_resource)

    def test_parse_from_resource(self):
        p = self.registry.get(self.resource['metadata']['name'])
        self.assertIsInstance(p, self.provider_class)

    def test_name_equals_resource_name(self):
        p = self.registry.get(self.resource['metadata']['name'])
        self.assertEqual(p.name, self.resource['metadata']['name'])

    def test_description_equals_resource_description(self):
        p = self.registry.get(self.resource['metadata']['name'])
        self.assertEqual(p.description, self.resource['spec']['description'])



# pylint: skip-file
"""Declares :class:`BaseLoader`."""
import abc


class BaseLoader:
    """The base class for all identity provider configuration loaders."""

    @abc.abstractmethod
    def load(self):
        """Loads identity provider configurations from specific sources."""
        raise NotImplementedError



# pylint: skip-file
import unittest

import ioc
import unimatrix.lib.test
from authlib.integrations.django_client import OAuth

from ..environment import EnvironmentLoader


@unimatrix.lib.test.unit
class EnvironmentLoaderTestCase(unittest.TestCase):

    def setUp(self):
        ioc.provide('OAuth2Registry', OAuth(), force=True)

    def test_load_from_environment(self):
        loader = EnvironmentLoader()
        providers = loader.load({
            'GOOGLE_OAUTH_CLIENT_ID': "foo",
            'GOOGLE_OAUTH_CLIENT_SECRET': "bar"
        })
        self.assertTrue(len(providers) == 1)

    def test_load_from_environment_with_empty(self):
        loader = EnvironmentLoader()
        providers = loader.load({})
        self.assertEqual(len(providers), 0)



# pylint: skip-file
from unittest import TestCase

import unimatrix.lib.test
from django.apps import apps

from ..subject import SubjectRepository


@unimatrix.lib.test.unit
class SubjectRepositoryTestCase(TestCase):

    @property
    def model(self):
        return apps.get_model('idc.AssertedSubject')

    def setUp(self):
        self.repo = SubjectRepository()

    def test_model_is_correct(self):
        self.assertEqual(self.repo.model, self.model)

    def test_do_not_exist_is_proxied(self):
        self.assertEqual(self.repo.DoesNotExist, self.model.DoesNotExist)



"""The :mod:`iam.ext.idc.lib.oauth2` implements OAuth2."""
from collections import defaultdict

import ioc

from .providers import GenericOAuth2Provider
from .providers import OIDCGenericProvider
from .providers import GoogleOIDCProvider
from .providers import MockOAuth2Provider


PROVIDER_CLASSES = defaultdict(lambda: GenericOAuth2Provider, {
    'accounts.google.com': GoogleOIDCProvider,
    'mock': MockOAuth2Provider
})


@ioc.inject('oauth2', 'OAuth2Registry')
def get_oauth2_provider(resource, oauth2):
    """Return an OAuth2 provider implementation class based on the type
    member of `resource.
    """
    return PROVIDER_CLASSES[ resource.get('type') ].fromresource(resource, oauth2)



# pylint: skip-file
import json
import os
import tempfile
import unittest

import ioc
import unimatrix.lib.test
from django.urls import reverse
from django.test import TestCase

from iam.ext.idc.lib import SecretsDirectoryLoader
from iam.ext.idc.infra.repo import SubjectRepository


TEST_PROVIDERS = """
---
apiVersion: v1alpha1
kind: OAuth2Provider
metadata:
  name: oauth-oidc
type: accounts.google.com
spec:
  description: Google
  credentials:
    id: foo
    secret: bar
  scope:
  - openid
  - email
  - profile


---
apiVersion: v1alpha1
kind: OAuth2Provider
metadata:
  name: default
type: accounts.google.com
spec:
  description: Google
  defaultProvider: true
  credentials:
    id: foo
    secret: bar
  scope:
  - openid
  - email
  - profile


---
apiVersion: v1alpha1
kind: OAuth2Provider
metadata:
  name: mock
  annotations:
    lsid: "1"
type: mock
spec:
  description: Mock Provider for Authenticated User
  credentials:
    id: foo
    secret: bar
  scope:
  - openid
  - email
  - profile


---
apiVersion: v1alpha1
kind: OAuth2Provider
metadata:
  name: mock-nouser
type: mock
spec:
  description: Mock Provider for Authenticated User
  credentials:
    id: foo
    secret: bar
  scope:
  - openid
  - email
  - profile
"""


@unittest.skipIf(not os.getenv('DJANGO_SETTINGS_MODULE'),
    "AuthenticationEndpointsTestCase requires DJANGO_SETTINGS_MODULE")
@unimatrix.lib.test.system
class AuthenticationEndpointsTestCase(TestCase):
    fixtures = ['test-subjects']

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        secdir = os.environ['APP_SECDIR'] = tempfile.mkdtemp()
        providers_dir = os.path.join(secdir, 'iam.unimatrixone.io/providers')
        os.makedirs(providers_dir)
        with open(os.path.join(providers_dir, 'test.yml'), 'w') as f:
            f.write(TEST_PROVIDERS)
        ioc.provide('iam.IdentityProviderLoaders', [
            SecretsDirectoryLoader(secdir=secdir)
        ], force=True)

    def test_authorize_logs_in_subject(self):
        response = self.client.get(
            reverse('iam:authorize', kwargs={'using': 'mock'}),
            follow=True)
        self.assertEqual(response.status_code, 200, response)

    def test_authorize_logs_not_in_none_subject(self):
        response = self.client.get(
            reverse('iam:authorize', kwargs={'using': 'mock-nouser'}),
            follow=True)
        self.assertEqual(response.status_code, 403)

    def test_login_sets_redirect_cookie(self):
        response = self.client.get(
            reverse('iam:login', kwargs={'using': 'mock'})\
                + '?next=/hello'
        )
        self.assertIn('idc.post_login_redirect', response.cookies)

    def test_login_redirects(self):
        response = self.client.get(
            reverse('iam:login', kwargs={'using': 'mock'})\
                + '?next=/hello',
            follow=True
        )
        self.assertNotIn('idc.post_login_redirect', response.cookies)
        self.assertEqual(response.request['PATH_INFO'], '/hello')

    def test_login_unknown_provider_returns_404(self):
        response = self.client.get(reverse('iam:login', kwargs={'using': 'bla'}))
        self.assertEqual(response.status_code, 404)

    def test_login_default_redirects(self):
        response = self.client.get(reverse('iam:login.default'))
        self.assertEqual(response.status_code, 302)

    def test_oidc_login_redirects(self):
        response = self.client.get(
            reverse('iam:login', kwargs={'using': 'oauth-oidc'}))
        self.assertEqual(response.status_code, 302)

    def test_oidc_invalid_provider_response(self):
        response = self.client.get(reverse('iam:authorize', kwargs={'using': 'oauth-oidc'}))
        self.assertEqual(response.status_code, 500)

    def test_get_default_provider_returns_default(self):
        response = self.client.get(reverse('iam:providers.default'))
        dto = json.loads(bytes.decode(response.content))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(dto['name'], 'default')

    def test_list_providers(self):
        response = self.client.get(reverse('iam:providers'))
        self.assertEqual(response.status_code, 200)



# pylint: disable=arguments-differ
"""Declares :class:`OIDCAuthenticationBackend`."""
import ioc

from django.contrib.auth.backends import BaseBackend


class OIDCAuthenticationBackend(BaseBackend):
    """Implements Django authentication for OAuth2 providers using
    OpenID Connect.
    """

    def authenticate(self, request, token=None, provider=None):  # pragma: no cover
        """Use `token` to receive a claims set from the OIDC provider
        `provider`.
        """
        return provider.token_to_subject(request, token)

    @ioc.inject('subjects', 'iam.SubjectRepository') # pragma: no cover
    def get_user(self, lsid, subjects):
        """Return the :term:`Subject` identified by `lsid`."""
        return subjects.get(pk=lsid)



# pylint: disable=R0801
"""Declares :class:`ResourceSchema`."""
import marshmallow
import marshmallow.fields

from .objectmeta import ObjectMetaSchema


class ResourceSchema(marshmallow.Schema):
    """Declares a schema for identity provider resources."""
    apiVersion = marshmallow.fields.String(
        required=True
    )

    kind = marshmallow.fields.String(
        required=True
    )

    metadata = marshmallow.fields.Nested(
        ObjectMetaSchema,
        required=True
    )

    spec = marshmallow.fields.Dict(
        required=True
    )

    type = marshmallow.fields.String(
        required=False
    )



# Generated by Django 3.1.1 on 2020-09-09 00:31

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AssertedSubject',
            fields=[
                ('lsid', models.BigAutoField(db_column='lsid', editable=False, primary_key=True, serialize=False, verbose_name='Local Subject ID')),
                ('asid', models.CharField(db_column='asid', max_length=254, unique=True, verbose_name='Asserted Identity')),
                ('external', models.BooleanField(db_column='is_external', default=True, verbose_name='External')),
                ('is_superuser', models.BooleanField(db_column='is_superuser', default=False, verbose_name='Superuser')),
            ],
            options={
                'verbose_name': 'Asserted Subject',
                'verbose_name_plural': 'Asserted Subjects',
                'db_table': 'assertedsubjects',
                'default_permissions': [],
            },
        ),
        migrations.AddConstraint(
            model_name='assertedsubject',
            constraint=models.CheckConstraint(check=models.Q(('external', True), ('is_superuser', True), _negated=True), name='assertedsubjects_not_superuser_and_external'),
        ),
    ]



# pylint: disable=invalid-name
"""Declares :class:`BaseOAuth2Provider`."""
import ioc

from iam.canon.idc import OAuth2ProviderSpecSchema
from ...provider import BaseProvider


class BaseOAuth2Provider(BaseProvider):
    """The base class for all OAuth2 providers."""
    schema = OAuth2ProviderSpecSchema()
    metadata_url = None

    @property
    def name(self):
        """Return the name with which the client was registered in the
        :mod:`authlib` library.
        """
        return self._client.name

    @property
    def description(self):
        """Return the description as specified in the resource."""
        return self._spec['description']

    @classmethod
    def fromresource(cls, resource, oauth, *args, **kwargs): # pylint: disable=arguments-differ
        """Create a new :class:`OIDCGenericProvider` from a provider
        resource.
        """
        meta = resource['metadata']
        spec = cls.schema.load(resource['spec'])
        credentials = spec.pop('credentials')
        client = cls.get_client(oauth, meta, spec, credentials,
            url=getattr(cls, 'metadata_url', None),
            client_kwargs=cls.get_client_kwargs(spec))
        return cls.fromclient(client, spec)

    @staticmethod
    def get_client(oauth, meta, spec, credentials, url=None, client_kwargs=None): # pylint: disable=R0913
        """Create an OAuth2 client using the provided parameters."""
        url = url or spec.get('url')
        kwargs = {
            'name': meta['name'],
            'client_id': credentials['id'],
            'client_secret': credentials['secret'],
            'client_kwargs': client_kwargs
        }
        if url:
            kwargs['server_metadata_url'] = url
        elif spec.get('metadata'):
            kwargs.update({
                'request_token_url': spec['metadata']['request_token']['url'],
                'request_token_params': spec['metadata']['request_token']['params'],
                'access_token_url': spec['metadata']['access_token']['url'],
                'access_token_params': spec['metadata']['access_token']['params'],
                'base_url': spec['metadata']['base_url'],
                'authorize_url': spec['metadata']['authorize_url']
            })
        else:
            raise ValueError("Specify either providerMetadata or metadataUrl")
        return oauth.register(**kwargs)

    @staticmethod
    def get_client_kwargs(*args, **kwargs): # pylint: disable=unused-argument
        """Return the client keyword arguments for the specified
        provider `spec`.
        """
        return None

    @classmethod
    def fromclient(cls, client, spec):
        """Return a new :class:`BaseOAuth2Provider` from a :mod:`authlib`
        OAuth2 client object.
        """
        return cls(client, spec)

    def __init__(self, client, spec, *args, **kwargs):
        super().__init__(spec, *args, **kwargs)
        self._client = client

    def authorize_redirect(self, request, redirect_to, **kwargs):
        """Initiate the OAuth2 login dialog by redirecting the user to the
        SSO provider login screen.
        """
        return self._client.authorize_redirect(request, redirect_to, **kwargs)

    start_authentication = authorize_redirect

    def get_asserted_subject(self, request):
        """Invoked after the OAuth2 provider redirects the client upon
        succesfully completing the authentication dialog.
        """
        return self.authenticate(
            request,
            token=self._client.authorize_access_token(request),
            provider=self
        )

    def get_asserted_subject_id(self, cs): # pylint: disable=invalid-name
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        raise NotImplementedError("Subclasses must override this method.")

    @ioc.inject('subjects', 'iam.SubjectRepository')
    @ioc.inject('factory', 'iam.SubjectFactory')
    def token_to_subject(self, request, token, subjects, factory):
        """Exchange an access token for an OIDC ID token, and get or
        create a corresponding :term:`Subject` from the persistent
        data store.
        """
        cs = self._client.parse_id_token(request, token)
        asid = self.get_asserted_subject_id(cs)
        try:
            sub = subjects.get(asid=asid)
        except subjects.DoesNotExist:
            sub = factory.new()
        sub.asid = asid
        sub.external = bool(self._spec.get('external'))
        sub.is_superuser = self.is_superuser(cs)
        subjects.persist(sub)
        return sub

    def is_superuser(self, cs):
        """Return a boolean indicating if the :term:`Subject` asserted by
        the claims set is a superuser.
        """
        yes = False
        for superuser in (self._spec.get('superusers') or []):
            claim, value = str.split(superuser, ':')
            if cs.get(claim) != value:
                continue
            yes = True
            break
        return yes



# pylint: skip-file
"""Declares :class:`MockOAuth2Provider`."""
import ioc
from django.http import HttpResponseRedirect

from .generic import GenericOAuth2Provider


class MockOAuth2Provider(GenericOAuth2Provider):

    @property
    def name(self):
        return self._metadata['name']

    @property
    def description(self):
        return self._spec['description']

    @classmethod
    def fromresource(cls, resource, *args, **kwargs):
        return cls(resource['metadata'], cls.schema.load(resource['spec']))

    def __init__(self, metadata, spec):
        self._metadata = metadata
        self._spec = spec

    @ioc.inject('subjects', 'iam.SubjectRepository')
    def get_asserted_subject(self, request, subjects):
        annotations = self._metadata.get('annotations') or {}
        return subjects.get(pk=annotations['lsid'])\
            if annotations.get('lsid')\
            else None

    def start_authentication(self, request, redirect_to):
        return HttpResponseRedirect(redirect_to)



# pylint: skip-file
import os

import unimatrix.lib.test
from django.test import TestCase

from ..subject import SubjectRepository


@unimatrix.lib.test.system
class SubjectRepositoryTestCase(TestCase):

    def get_instance(self, **kwargs):
        return self.repo.model(asid=bytes.hex(os.urandom(16)))

    def setUp(self):
        self.repo = SubjectRepository()

    def test_exists_returns_false_on_empty_database(self):
        self.assertFalse(self.repo.exists(1))

    def test_persist(self):
        obj = self.get_instance()
        self.repo.persist(obj)
        self.assertTrue(self.repo.exists(obj.pk))

    def test_get_by_pk(self):
        obj = self.get_instance()
        self.repo.persist(obj)
        self.repo.get(pk=obj.pk)
